/*----------------------------------------------------------------------------*-

  ttrd2-06a-t0405a-v001a_scheduler.c (Release 2017-08-18a)

  ----------------------------------------------------------------------------

  Time-Triggered Co-operative (TTC) task scheduler for STM32F405.

  Designed to run on DuplicaTTor Evaluation Board (v002).

  Provides support for Backup Tasks.

  See "ERES2" (Chapter 6) for further information about this simple scheduler.

-*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*-

  This code is copyright (c) 2014-2016 SafeTTy Systems Ltd.

  This code forms part of a Time-Triggered Reference Design (TTRD) that is 
  documented in the following book: 

   Pont, M.J. (2016) 
   "The Engineering of Reliable Embedded Systems (Second Edition)", 
   Published by SafeTTy Systems Ltd. ISBN: 978-0-9930355-3-1.

  Both the TTRDs and the above book ("ERES2") describe patented 
  technology and are subject to copyright and other restrictions.

  This code may be used without charge: [i] by universities and colleges on 
  courses for which a degree up to and including 'MSc' level (or equivalent) 
  is awarded; [ii] for non-commercial projects carried out by individuals 
  and hobbyists.

  Any and all other use of this code and / or the patented technology 
  described in ERES2 requires purchase of a ReliabiliTTy Technology Licence:
  http://www.safetty.net/technology/reliabilitty-technology-licences

  Please contact SafeTTy Systems Ltd if you require clarification of these 
  licensing arrangements: http://www.safetty.net/contact

  CorrelaTTor, DecomposiTTor, DuplicaTTor, MoniTTor, PredicTTor, ReliabiliTTy,  
  SafeTTy, SafeTTy Systems, TriplicaTTor and WarranTTor are registered 
  trademarks or trademarks of SafeTTy Systems Ltd in the UK & other countries.

-*----------------------------------------------------------------------------*/

// Processor Header
#include "..\main\main.h"

// ------ Private variable definitions ---------------------------------------

// The array of tasks
// Check array size in scheduler header file
static sTask_t SCH_tasks_g[SCH_MAX_TASKS];

// The array of backup tasks
static sTask_t SCH_backup_tasks_g[SCH_MAX_TASKS];

// Tick count
static uint32_t Tick_count_g = 0;

// ------ Private function prototypes ----------------------------------------

void SysTick_Handler(void);

/*----------------------------------------------------------------------------*-

  SCH_Init_Milliseconds()

  Scheduler initialisation function.  Prepares scheduler data structures and 
   sets up timer interrupts every TICKms milliseconds.

  You must call this function before using the scheduler.

  [Required_SystemCoreClock frequency can be found in main.h.]

  PARAMETERS:
     TICKms is the required Tick Interval (milliseconds)

  LONG-TERM DATA:
     SCH_tasks_g (W)

  MCU HARDWARE:
     SysTick timer.

  PRE-CONDITION CHECKS:
     None.

  POST-CONDITION CHECKS:
     None.

  ERROR DETECTION / ERROR HANDLING:
     PROCESSOR_Perform_Safe_Shutdown() is called if the requested timing 
     requirements cannot be met.

  RETURN VALUE:
     None.

-*----------------------------------------------------------------------------*/
void SCH_Init_Milliseconds(const uint32_t TICKms)
   {
   for (uint32_t Task_id = 0; Task_id < SCH_MAX_TASKS; Task_id++)
      {
      // Set pTask to "null pointer"
      SCH_tasks_g[Task_id].pTask = SCH_NULL_PTR;
      }

   // Using CMSIS
      
   // SystemCoreClock gives the system operating frequency (in Hz)
   if (SystemCoreClock != REQUIRED_PROCESSOR_CORE_CLOCK)
      {
      // Fatal error
      PROCESSOR_Perform_Safe_Shutdown();
      }

   // Now to set up SysTick timer for "ticks" at interval TICKms
   if (SysTick_Config(TICKms * SystemCoreClock / 1000))
      {
      // Fatal error
      PROCESSOR_Perform_Safe_Shutdown();
      }

   // Timer is started by SysTick_Config():
   // we need to disable SysTick timer and SysTick interrupt until
   // all tasks have been added to the schedule.
   SysTick->CTRL &= 0xFFFFFFFC;
   }

/*----------------------------------------------------------------------------*-

  SCH_Start()

  Starts the scheduler, by enabling SysTick interrupt.

  NOTES: 
  * All tasks must be added before starting scheduler.
  * Any other interrupts MUST be synchronised to this tick.  

  PARAMETERS:
     None.

  LONG-TERM DATA:
     None.

  MCU HARDWARE:
     SysTick timer.

  PRE-CONDITION CHECKS:
     None.

  POST-CONDITION CHECKS:
     None.

  ERROR DETECTION / ERROR HANDLING:
     None.

  RETURN VALUE:
     None.

-*----------------------------------------------------------------------------*/
void SCH_Start(void) 
   {
   // Enable SysTick timer
   SysTick->CTRL |= 0x01;

   // Enable SysTick interrupt
   SysTick->CTRL |= 0x02;
   }


/*----------------------------------------------------------------------------*-

  SysTick_Handler()

  [Function name determined by CMIS standard.]

  This is the scheduler ISR.  It is called at a rate 
  determined by the timer settings in the SCH_Init() function.

  PARAMETERS:
     None.

  LONG-TERM DATA:
     Tick_count_g (W)

  MCU HARDWARE:
     SysTick timer.

  PRE-CONDITION CHECKS:
     None.

  POST-CONDITION CHECKS:
     None.

  ERROR DETECTION / ERROR HANDLING:
     Checks Tick_count_g value => 'Fail Safe'
     [See ERES2, Ch2 for details.]

  RETURN VALUE:
     None.

-*----------------------------------------------------------------------------*/
void SysTick_Handler(void)
   {
   // Increment tick count and check against limit
   if (++Tick_count_g > SCH_TICK_COUNT_LIMIT)
      {
      // One or more tasks has taken too long to complete
      PROCESSOR_Perform_Safe_Shutdown();
      }
   } 

/*----------------------------------------------------------------------------*-

  SCH_Dispatch_Tasks()

  This is the 'dispatcher' function.  When a task (function)
  is due to run, SCH_Dispatch_Tasks() will run it.
  This function must be called (repeatedly) from the main loop.

  Supports Backup Tasks.

  Triggers move to "idle" mode when all tasks have been released.

  PARAMETERS:
     None.

  LONG-TERM DATA:
     SCH_tasks_g (W)
     Tick_count_g (W)

  MCU HARDWARE:
     Triggers move to idle mode.

  PRE-CONDITION CHECKS:
     None.

  POST-CONDITION CHECKS:
     None.

  ERROR DETECTION / ERROR HANDLING:
     None.

  RETURN VALUE:
     None.

-*----------------------------------------------------------------------------*/
void SCH_Dispatch_Tasks(void) 
   {
   uint32_t Status, Backup_status;

   __disable_irq();  
   uint32_t Update_required = (Tick_count_g > 0);  // Check tick count
   __enable_irq();

   while (Update_required)
      {
      // Go through the task array
      for (uint32_t Task_id = 0; Task_id < SCH_MAX_TASKS; Task_id++)
         {
         // Check if there is a task at this location
         if (SCH_tasks_g[Task_id].pTask != SCH_NULL_PTR)
            {
            if (--SCH_tasks_g[Task_id].Delay == 0)
               {
               Status = (*SCH_tasks_g[Task_id].pTask)();  // Run the task

               if (Status == RETURN_ABNORMAL_STATE)
                  {
                  // Fault detected during task execution
                  // Try to switch to a Backup Task
                  Backup_status = SCH_Replace_Task(Task_id);

                  if (Backup_status == RETURN_ABNORMAL_STATE)
                     {
                     // No Backup Task (or it has already been tried)
                     // Here we force a shutdown
                     PROCESSOR_Perform_Safe_Shutdown();
                     }
                  }

               // All tasks are periodic: schedule task to run again
               SCH_tasks_g[Task_id].Delay = SCH_tasks_g[Task_id].Period;
               }
            }         
         } 

      __disable_irq();
      Tick_count_g--;                       // Decrement the count
      Update_required = (Tick_count_g > 0); // Check again
      __enable_irq();
      }

   // The scheduler enters idle mode at this point 
   __WFI();     
   }

/*----------------------------------------------------------------------------*-

  SCH_Add_Task()

  Adds a task (function) to the schedule.
  The task will be released periodically by the scheduler.

  PARAMETERS:
     pTask  : The name of the task (function) to be scheduled.
              NOTE: All scheduled functions must be 'void, void' -
              that is, they must take no parameters, and have 
              a void return type (in this design).
                   
     DELAY  : The interval (ticks) before the task is first executed.

     PERIOD : Task period (in ticks).  Must be > 0.

  LONG-TERM DATA:
     SCH_tasks_g (W)
  
  MCU HARDWARE:
     None.

  PRE-CONDITION CHECKS:
     1. There is space in the task array.
     2. The task is periodic ('one-shot' tasks are not supported. 

  POST-CONDITION CHECKS:
     None.

  ERROR DETECTION / ERROR HANDLING:
     PROCESSOR_Perform_Safe_Shutdown() is called:
     - if the task cannot be added to the schedule (array too small)
     - if an attempt is made to schedule a "one shot" task

  RETURN VALUE:  
     None.
 
-*----------------------------------------------------------------------------*/
uint32_t SCH_Add_Task(uint32_t (* pTask)(),
                      const uint32_t DELAY,
                      const uint32_t PERIOD)
   {
   uint32_t Task_id = 0;
   
   // First find a gap in the array (if there is one)
   while ((SCH_tasks_g[Task_id].pTask != SCH_NULL_PTR) 
          && (Task_id < SCH_MAX_TASKS))
      {
      Task_id++;
      } 
   
   // Have we reached the end of the list?   
   if (Task_id == SCH_MAX_TASKS)
      {
      // Task list is full - fatal error
      PROCESSOR_Perform_Safe_Shutdown();
      }
      
   // Check for "one shot" tasks
   if (PERIOD == 0)
      {
      // We do not allow "one shot" tasks (all tasks must be periodic)
      PROCESSOR_Perform_Safe_Shutdown();
      }

   // If we're here, there is a space in the task array
   // and the task to be added is periodic
   SCH_tasks_g[Task_id].pTask  = pTask;

   SCH_tasks_g[Task_id].Delay  = DELAY + 1;
   SCH_tasks_g[Task_id].Period = PERIOD;

   return Task_id;
   }

/*--------------------------------------------------------------------*-

  SCH_Add_Backup_Task()

  Adds a Backup Task to the schedule.

  The Backup Task will replace the corresponding Main Task if
  a fault is detected in the Main Task.

  pTask  - The name of the task (function) to be scheduled.
           NOTE: All scheduled functions must be
           'uint32_t, const uint32_t'

 [DELAY  - Same as the Main Task]

 [PERIOD - Same as the Main Task.]

  RETURN VALUE: None.

  Note: Parameters are not checked!

-*--------------------------------------------------------------------*/
void SCH_Add_Backup_Task(const uint32_t INDEX,
                         uint32_t (* pTask)())
   {
   SCH_backup_tasks_g[INDEX].pTask  = pTask;

   SCH_backup_tasks_g[INDEX].Delay  = SCH_tasks_g[INDEX].Delay;
   SCH_backup_tasks_g[INDEX].Period = SCH_tasks_g[INDEX].Period;
   }

/*--------------------------------------------------------------------*-

  SCH_Replace_Task()

  Aim to replace Main Tasks with a corresponding Backup Task, if:
  [i]  a Backup Task is available, and
  [ii] any Backup Task is not already in use

  If there is no Backup Task (or this is already in use),
  we return a fault code (and expect the system to Fail Silent).

  RETURN VALUE:  RETURN_ABNORMAL_STATE or RETURN_NORMAL_STATE

-*--------------------------------------------------------------------*/
uint32_t SCH_Replace_Task(const uint32_t TASK_INDEX)
   {
   uint32_t Return_value = RETURN_NORMAL_STATE;

   if (SCH_tasks_g[TASK_INDEX].pTask == 0)
      {
      // No task at this location
      // Serious problem!

      // Return a fault code
      Return_value = RETURN_ABNORMAL_STATE;
      }

   if (Return_value == RETURN_NORMAL_STATE)
      {
      if (SCH_backup_tasks_g[TASK_INDEX].pTask == 0)
         {
         // No Backup Task at this location

         // Here we treat this as a fault
         // (other behaviour may make more sense
         // in your application)
         Return_value = RETURN_ABNORMAL_STATE;
         }
      }

   if (Return_value == RETURN_NORMAL_STATE)
      {
      // If we are here, there is a Backup Task

      // Check if the Backup Task is already in use
      if (SCH_backup_tasks_g[TASK_INDEX].pTask ==
          SCH_tasks_g[TASK_INDEX].pTask)
         {
         // We are already using the Backup Task
         // Here we treat this as a fault
         // (other behaviour may make more sense
         // in your application)
         Return_value = RETURN_ABNORMAL_STATE;
         }
      }

   if (Return_value == RETURN_NORMAL_STATE)
      {
      // If we are here, there is a Backup Task
      // and it is not currently being used

      // Use the Backup Task
      SCH_tasks_g[TASK_INDEX].pTask
      = SCH_backup_tasks_g[TASK_INDEX].pTask;

      Return_value = RETURN_NORMAL_STATE;
      }

   // Single return point
   return Return_value;
   }

/*----------------------------------------------------------------------------*-
  ------------------------------ END OF FILE ---------------------------------
-*----------------------------------------------------------------------------*/
